#!/usr/bin/python

import os
import re
import optparse
import sys
import random
import copy
import curses
import curses.textpad
import md5
import platform
try:
    import xmlrpclib
except ImportError:
    print "This script needs the xmlrpclib module"
    sys.exit(1)

# URL for when this is called by a wiki
wiki_url = "http://imonk.com/issue"
wiki_write_access = "http://issue_py:issue_py@imonk.com/issue"

prefix = '['
suffix = ']'

wiki_id = ""

key_color_map = {
	'active'	:	'#00dd00',
	'dormant'	:	'#777777',
	'pending'	:	'#dd9922',
	'verify'	:	'#0000dd',
	'none'		:	'#dd0000',
	}

re_kv = re.compile(r'\%s([^=\n]*)=([^\%s\%s\n]*)\%s' % (prefix, prefix, suffix, suffix))
re_has = re.compile(r'([^~]*)~(.*)', re.DOTALL)
re_nothas = re.compile(r'([^\?]*)\?(.*)', re.DOTALL)
re_eq = re.compile(r'([^=]*)=(.*)', re.DOTALL)
re_neq = re.compile(r'([^\^]*)\^(.*)', re.DOTALL)
re_lt = re.compile(r'([^<]*)<(.*)', re.DOTALL)
re_gt = re.compile(r'([^>]*)>(.*)', re.DOTALL)
re_command = re.compile(r'([^:]*):(.*)', re.DOTALL)
re_first_num = re.compile(r'[^0-9\+\-]*(\+|\-)?(\d+).*', re.DOTALL)
re_first_num_str = re.compile(r'[^0-9\+\-]*(\+|\-)?(\d+)s', re.DOTALL)
re_latex_kv = re.compile(r'%([^=]*)=(\S*)', re.DOTALL)

def filter(filename):
	if filename[-4:] == ".swp":
		return 1
	return 0

class StringCompare:
	def __init__(self, key):
		self.key = key
	def __call__(self, fa, fb):
		if fa.attr(self.key) == None and fb.attr(self.key) == None:
			return 0

		if fa.attr(self.key) == None:
			return 1

		if fb.attr(self.key) == None:
			return -1


		a = fa.attr(self.key)
		b = fb.attr(self.key)

		if a < b:
			return 1
		if a > b:
			return -1
		return 0

class NumberCompare:
	def __init__(self, key):
		self.key = key
	def __call__(self, fa, fb):

		if fa.attr(self.key) == None and fb.attr(self.key) == None:
			return 0

		if fa.attr(self.key) == None:
			return 1

		if fb.attr(self.key) == None:
			return -1


		try:
			a = float(fa.attr(self.key))
		except:
			return 1

		try:
			b = float(fb.attr(self.key))
		except:
			return -1

		if a < b:
			return 1
		if a > b:
			return -1
		return 0

class RevNumberCompare:
	def __init__(self, key):
		self.compare = NumberCompare(key)
	def __call__(self, fa, fb):
		return -self.compare(fa,fb)

class RevStringCompare:
	def __init__(self, key):
		self.compare = StringCompare(key)
	def __call__(self, fa, fb):
		return -self.compare(fa,fb)

def ltx(s):
	if s == None:
		return ""
	s = s.replace('_', '\_')
	s = s.replace('&', '\&')
	s = s.replace('>', '$>$')
	s = s.replace('<', '$<$')
	return s


class Set:
	def __init__(self):
		self.map = {}
		self.files = []
		self.terse = 0
	def edit(self):
		for f in self.files:
			fp = open(f.filename)
			self.txt_buffer = fp.read()
			fp.close()
			txt = curses.wrapper(self.cedit)
			fp = open(f.filename, 'w')
			fp.write(txt)
			fp.close()
	def cedit(self,stdscr):
		stdscr.addstr(self.txt_buffer)
		tbox = curses.textpad.Textbox(stdscr)
		return tbox.edit()
	def add(self, file):
		self.map[file.filename] = file
		self.files.append(file)
	def join(self, set):
		for f in set.files:
			if not self.map.has_key(f.filename):
				self.add(f)
	def intersect(self, set):
		for k in self.map.keys():
			if not set.map.has_key(k):
				del self.map[k]
		self.rebuild()
	def load(self, dir_root):
		if os.path.isfile(dir_root):
			if filter(dir_root):
				return
			file =  File(dir_root)
			if file.process():
				self.add(file)
			return
		
		for root, dirs, files in os.walk(dir_root, topdown=True):
			if '.svn' in dirs:
				dirs.remove('.svn')

			for d in dirs:
				if os.path.islink(d):
					self.load(d)

			for f in files:
				filename = os.path.join(root, f)


				if filter(filename):
					continue

				file =  File(filename)
				if file.process():
					self.add(file)
	def rebuild(self):
		self.files = []
		for k in self.map.keys():
			self.files.append(self.map[k])
	def cap(self, count):
		count = int(float(count))
		rm_list = self.files[count:]
		for f in rm_list:
			del self.map[f.filename]
		self.files = self.files[:count]
	def sort(self, key, reverse=False):
		if reverse:
			self.files.sort(RevNumberCompare(key))
		else:
			self.files.sort(NumberCompare(key))
	def ssort(self, key, reverse=False):
		if reverse:
			self.files.sort(RevStringCompare(key))
		else:
			self.files.sort(StringCompare(key))
	def set(self, key_value):
		kv = key_value.split(':')
		if len(kv) != 2:
			return

		for f in self.files:
			f.map[kv[0].strip()] = kv[1].strip()
			f.dirty = 1

	def setkv(self, key, value):
		for f in self.files:
			f.map[key.strip()] = value.strip()
			f.dirty = 1

	def incr(self, key_value):
		kv = key_value.split(':')
		if len(kv) != 2:
			return

		val = float(kv[1].strip())
		key = kv[0].strip()
		for f in self.files:
			if f.map.has_key(key):
				f.map[key] = str(float(f.map[key]) + val)
				f.dirty = 1

	def delete(self, key):
		for f in self.files:
			if f.map.has_key(key):
				del f.map[key]
				f.del_keys.append(key)
				f.dirty = 1

	def push(self):
		for f in self.files:
			f.push()

	def new(self, dir):
		n = 0

		(head,tail) = os.path.split(dir)

		if dir == '.':
			prefix = ''
		elif head == '':
			dir = '.'
			prefix = tail
		else:
			dir = head
			prefix = tail

		re_autonamed = re.compile(r'%s(\d+)' % (prefix), re.DOTALL)


		for f in os.listdir(dir):
			m = re_autonamed.match(f)
			if m:
				id = int(m.group(1))
				if id > n:
					n = id
		n += 1

		filename =  os.path.join(dir,"%s%d" % (prefix,n))
		txt = ""
		fp = open(filename, "w")
		fp.write(txt)
		fp.close()


		file =  File(filename)
		if file.process():
			self.add(file)

		self.files = [ file ]

	def filterExists(self, key):
		for k in self.map.keys():
			if not self.map[k].attr(key):
				del self.map[k]
		self.rebuild()
	def filterHas(self, key, value):
		for k in self.map.keys():
			pattern = ".*%s.*" % (value)
			re_m = re.compile(pattern)
			v = self.map[k].attr(key)
			try:
				if v == None or not re_m.match(v):
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def filterNotHas(self, key, value):
		for k in self.map.keys():
			pattern = ".*%s.*" % (value)
			re_m = re.compile(pattern)
			v = self.map[k].attr(key)
			try:
				if v != None and re_m.match(v):
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def filterEqual(self, key, value):
		for k in self.map.keys():
			v = self.map[k].attr(key)
			try:
				if v == None or v.strip() != value.strip():
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def filterNumEqual(self, key, value):
		for k in self.map.keys():
			v = self.map[k].attr(key)
			try:
				if v == None or float(v.strip()) != float(value.strip()):
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def filterAttrEqual(self, key1, key2):
		for k in self.map.keys():
			v1 = self.map[k].attr(key1)
			v2 = self.map[k].attr(key2)
			try:
				if v1 == None or v2 == None or v1.strip() != v2.strip():
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def filterNotEqual(self, key, value):
		for k in self.map.keys():
			v = self.map[k].attr(key)
			try:
				if v == None or v.strip() == value.strip():
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def filterLessThan(self, key, value):
		for k in self.map.keys():
			v = self.map[k].attr(key)
			try:
				if v == None or float(v.strip()) >= float(value.strip()):
					del self.map[k]
			except:
					print k
					del self.map[k]
		self.rebuild()
	def filterLessThanEq(self, key, value):
		for k in self.map.keys():
			v = self.map[k].attr(key)
			try:
				if v == None or float(v.strip()) > float(value.strip()):
					del self.map[k]
			except:
					print k
					del self.map[k]
		self.rebuild()
	def filterGreaterThan(self, key, value):
		for k in self.map.keys():
			v = self.map[k].attr(key)
			try:
				if v == None or float(v.strip()) <= float(value.strip()):
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def filterGreaterThanEq(self, key, value):
		for k in self.map.keys():
			v = self.map[k].attr(key)
			try:
				if v == None or float(v.strip()) < float(value.strip()):
					del self.map[k]
			except:
					del self.map[k]
		self.rebuild()
	def buildDependencies(self):
		# clear dependencies
		for f in self.files:
			f.depend_on = []
			f.dependent = []
		# build dependencies
		for f in self.files:
			if f.attr('dependencies'):
				deps = f.attr('dependencies').strip().split(',')
				for dep in deps:
					full_dep = os.path.join(os.path.dirname(f.filename), dep)
					local_dep = os.path.join(".", dep)
					if self.map.has_key(full_dep):
						other = self.map[full_dep]
						f.depend_on.append(other)
						other.dependent.append(f)
					elif self.map.has_key(local_dep):
						other = self.map[local_dep]
						f.depend_on.append(other)
						other.dependent.append(f)
	def addDeps(self, file, txt, depth):
		max = 20
		file.visit = 1
		if depth == 0:
			color = 'peachpuff'
		elif file.attr('status') == 'active':
			color = 'yellow'
		elif len(file.depend_on) == 0:
			color = 'palegreen'
		else:
			color = 'grey80'
		txt += '"%s (%s)" [shape=box style=filled color=%s];\n' % (file.attr('title')[:max], file.filename, color)
		for dep in file.depend_on:
			if dep.visit == 0:
				txt = self.addDeps(dep, txt, depth+1)
			txt += '"%s (%s)" -> "%s (%s)";\n' % (file.attr('title')[:max], file.filename, dep.attr('title')[:max], dep.filename)

		return txt
				
	def dotDependencies(self, id):
		txt = 'digraph G {\n'
		txt += 'overlap=false;\n'
		txt += 'rankdir=LR;\n'

		for f in self.files:
			f.visit = 0

		try:
			file = self.map[os.path.join('.', id)]
			txt = self.addDeps(file, txt, 0)
		except:
			pass

		txt += '}\n'
		fp = open('%s.dot' % (id), 'w')
		fp.write(txt)
		fp.close()
		os.system('dot -Tps %s.dot -o %s.ps' % (id, id))
	def dotDependenciesAll(self):


		txt = 'digraph G {\n'
		txt += 'overlap=false;\n'
		txt += 'epsilon=0.86;\n'
		txt += 'rankdir=LR;\n'

		for f in self.files:
			f.visit = 0

		for f in self.files:
			if len(f.dependent) == 0:
				continue
			if f.visit == 0:
				txt = self.addDeps(f, txt, 0)

		txt += '}\n'

		id = 'all'
		fp = open('%s.dot' % (id), 'w')
		fp.write(txt)
		fp.close()
		os.system('dot -Tps %s.dot -v -o %s.ps' % (id, id))

	def latex(self, filename):
		self.buildDependencies()

		sections = {}
		attributes = []
		tfile = open(filename)
		template_txt = tfile.read()
		tfile.close()

		kvs = re_latex_kv.findall(template_txt)

		section_name = 'category'
		for kv in kvs:
			key = kv[0].strip()
			value = kv[1].strip()
			if key == 'section':
				section_name = value
			if key == 'attribute':
				attributes.append(value)

		kv_matches = re_latex_kv.finditer(template_txt)
		kvms = []
		for kvm in kv_matches:
			kvms.append(kvm)
		kvms.reverse()
		for kvm in kvms:
			if kvm.group(1) == 'depgraph':
				id = kvm.group(2)
				if id == 'all':
					self.dotDependenciesAll()
					tmp_txt = template_txt[:kvm.start()]
					tmp_txt += """
						\\begin{figure*}
						\centerline{ \includegraphics[width=5.5in]{%s.ps} }
						\label{overView}
						\end{figure*} """ % ('all')
					tmp_txt += template_txt[kvm.end():]
					template_txt = tmp_txt
				else:
					filename = os.path.join('.', id)
					self.dotDependencies(id)
					try:
						file = self.map[filename]
						tmp_txt = template_txt[:kvm.start()]
						tmp_txt += """
							\\begin{figure*}
							\centerline{ \includegraphics[width=7.0in]{%s.ps} }
							\caption{%s}
							\label{overView}
							\end{figure*} """ % (id, file.attr('title'))
						tmp_txt += template_txt[kvm.end():]
						template_txt = tmp_txt
					except:
						pass

		for f in self.files:
			section = f.attr(section_name)
			if not sections.has_key(section):
				sections[section] = []
			sections[section].append(f)

		for section in sections.keys():
			section_txt =  '\section{%s}\n\n' % (ltx(section))
			for f in sections[section]:
				section_txt += '\subsection{%s}\n\n' % (ltx(f.attr("title")))

				if len(attributes) > 0:
					section_txt += '\\begin{tabular*}{0.45\\textwidth}{| l | @{\\extracolsep{\\fill}} l |}\n \\hline \n'

					for attribute in attributes:
						if attribute == 'filename':
							section_txt += '%s &%s \\\\ \\hline\n' % ('filename', f.filename)
						else:
							section_txt += '%s &%s \\\\ \\hline\n' % (attribute, f.attr(attribute))

					section_txt += '\\end{tabular*}\n\\newline\\vspace{1.0em}\\newline\n'
				
				if f.txt.strip() == "":
					section_txt += '.'
				section_txt += f.txt.strip()
				section_txt += '\n\n'
			template_txt = template_txt.replace('<%s>' % section, section_txt)
		fp = open('out.tex', 'w')
		fp.write(template_txt)
		fp.close()
		os.system('latex out.tex')
		os.system('latex out.tex')
		os.system('dvipdf out.dvi')
	def display(self, format=""):
		do_html = 0
		if format[:5] == "html ":
			format = format[5:]
			do_html = 1
		txt = ""
		if do_html:
			txt += '<table class="inline">'
		if format == "":
			format = "filename%-10s: | :priority%7.1f: | :status%10s: | :  :owner%10s: | :  :title%s"
		sn = ""
		if self.terse == 0:
			if do_html:
				txt += '<tr class="row0">'
			for c in format.split(':'):
				if do_html:
					txt += "<th>"
				fields = c.split('%')
				if len(fields) < 2:
					txt += c
					continue
				match = re_first_num.match(fields[1])
				if match:
					sz = int(match.group(2))
					sn = ""
					if match.group(1) != None:
						sn = match.group(1)
					fmt = "%%%s%d.%ds" % (sn,sz,sz)
					txt += fmt % fields[0]
				else:
					fmt = "%s"
					txt += fmt % fields[0]
				if do_html:
					txt += "</th>"
			txt += "\n"
			l = len(txt)
			if not do_html:
				for i in range(0, l - 1):
					txt += '-'
			txt += "\n"
			if do_html:
				txt += '</tr>'

		people = {}
		for f in self.files:
			if do_html:
				txt += '<tr class="row1">'
			for c in format.split(':'):
				if do_html:
					txt += '<td>'
				fields = c.split('%')
				if len(fields) < 2:
					txt += c
					continue
				key = fields[0].strip()
				attr = f.attr(key)

				if key == 'owner' or key == 'submitter':
					if attr != None:
						people[attr] = 1

				match = re_first_num_str.match(fields[1])
				sz = None
				if match:
					sz = int(match.group(2))
					sn = ""
					if match.group(1) != None:
						sn = match.group(1)
					fmt = "%%%s%d.%ds" % (sn,sz,sz)
				else:
					fmt = "%%%s" % (fields[1])

				match = re_first_num.match(fields[1])
				sz = None
				if match:
					sz = int(match.group(2))
					sn = ""
					if match.group(1) != None:
						sn = match.group(1)
					sfmt = "%%%s%d.%ds" % (sn,sz,sz)

				if sn == "+":
					attr = attr[-sz:]


				if attr == None:
					if do_html:
						txt += f.asForm(f.attr('wikiname'),key,"")
					else:
						try:
							txt += fmt % '-'
						except:
							txt += sfmt % '-'
				else:
					attr = attr.strip()
					if fields[1].strip()[-1] == 'f':
						try:
							attr = float(attr)
						except:
							fmt = sfmt
					if do_html:
						if key == 'wikiname':
							attr = f.asLink(attr,attr)
						elif key == 'title':
							if f.attr('wikiname'):
								attr = f.asEdit(f.attr('wikiname'),attr)
						else:
							fmt = "%s"
							#attr = f.asLink(str(key) + ":" + str(attr),str(attr))
							attr = f.asForm(f.attr('wikiname'),key,str(attr))
					txt += fmt % attr
				if do_html:
					txt += "</td>"
			if do_html:
				txt += "</tr>"

			txt += "\n"

		colormapping = {	'active'	:	'green:bold:blackbg',
							'verify'	:	'magenta:blackbg',
							'blocked'	:	'red:blackbg',
							'closed'	:	'white:blackbg',
							'dormant'	:	'cyan:blackbg',
							'pending'	:	'yellow:blackbg',
							'issue:'	:	'yellow:bold',
							'IMI:'		:	'magenta:bold',
							'fe:'		:	'red:bold',
							'fe/'		:	'red:bold',
							'doc:'		:	'cyan:bold',
							'fix:'		:	'red:bold',
							'buy:'		:	'green:bold',
							'dw:'		:	'blue:bold',
							'kfp:'		:	'white:bold',
							'kfp2:'		:	'white:bold',
							'shrek4:'	:	'green:bold',
							'fvwm:'		:	'magenta',
							'dw/'		:	'cyan:bold',
							'work/'		:	'magenta:bold',
							'personal/'	:	'blue:bold',
							'fex:'		:	'red',
							'orang'		:	'red:blackbg',
							'aweber'	:	'red:blackbg',
							'hoss'		:	'black:bold:greenbg',
							'baboon'	:	'cyan:blackbg',
							'none'		:	'blue:blackbg',
							'|'			:	'blue:bluebg'}

		ansimap = {	'bold'		:	1,
					'black'		:	30,
					'red'		:	31,
					'green'		:	32,
					'yellow'	:	33,
					'blue'		:	34,
					'magenta'	:	35,
					'cyan'		:	36,
					'white'		:	37,
					'blackbg'	:	40,
					'redbg'		:	41,
					'greenbg'	:	42,
					'yellowbg'	:	43,
					'bluebg'	:	44,
					'magentabg'	:	45,
					'cyanbg'	:	46,
					'whitebg'	:	47}
		htmlmap = {	}

		peoplecolors = (
							'green:blackbg',
							'magenta:blackbg',
							'cyan:blackbg',
							'white:blackbg'
							)

		for person in people.keys():
			if not colormapping.has_key(person.strip()):
				random.seed(person)
				i = random.randint(0,len(peoplecolors)-1)
				colormapping[person.strip()] = peoplecolors[i]

		if do_html:
			if self.terse == 0:
				for key in colormapping.keys():
					codes = colormapping[key]
					l = ''
					for codename in codes.split(':'):
						if htmlmap.has_key(codename.strip()):
							code = htmlmap[codename.strip()]
							l += '<span style="color:%s">' % (code)
					l += key.strip()
					for codename in codes.split(':'):
						if htmlmap.has_key(codename.strip()):
							l += '</span>'
					txt = txt.replace(key.strip(), l)
		else:
			if self.terse == 0:
				for key in colormapping.keys():
					codes = colormapping[key]
					l = ''
					for codename in codes.split(':'):
						code = ansimap[codename.strip()]
						l += '[%dm' % (code)
					l += key.strip()
					l += '[0m'
					txt = txt.replace(key.strip(), l)
		if do_html:
			txt += "</table>"
		sys.stdout.write(txt)




class File:
	def __init__(self, filename):
		self.filename = filename
		self.map = {}
		self.map['fullname'] = filename.strip()
		wikiname = filename.strip()
		wikiname = wikiname.replace('/', ':')
		if wikiname[-4:] == '.txt':
			wikiname = wikiname[:-4]
		wikiname = wikiname.replace('.', '_')
		if wikiname[0] == '_':
			wikiname = 'issue' + wikiname[1:]
		self.map['wikiname'] = wikiname
		self.map['filename'] = os.path.basename(filename.strip())
		self.dirty = 0
		self.del_keys = []
		self.txt = ""
		self.depend_on = []
		self.dependent = []
		self.first_line = ""
	def asLink(self, key, label):
		return '<a href="%s/doku.php?id=%s">%s</a>' % (wiki_url, key, label)
	def asEdit(self, key, label):
		return '<a href="%s/doku.php?id=%s&do=edit">%s</a>' % (wiki_url, key, label)
	def asForm(self,wikiname, attr, value):
		color = "black"
		if key_color_map.has_key(value):
			color = key_color_map[value]
		return '''<form action="%s/doku.php?id=%s&do=query" class="search" method="POST">
			<INPUT type="hidden" name="wikiname" value="%s">
			<INPUT type="hidden" name="attr" value="%s">
			<INPUT style="border: 0px; color:%s" type="text" size="10" name="value" value="%s">
			<INPUT type="hidden" name="wiki_id" value="%s">
			</form>
			''' % (wiki_url, wiki_id, wikiname, attr, color, value, wiki_id)
	def attr(self, key):
		if self.map.has_key(key):
			return self.map[key]
		else:
			if key == 'title':
				return self.first_line
			return None
	def process(self):
		try:
			fp = open(self.filename)
			self.first_line = fp.readline().strip().strip('=')
			fp.seek(0)
			file_txt = fp.read()
			fp.close()


			kvs = re_kv.findall(file_txt)
			self.txt = re_kv.sub("", file_txt)
			for kv in kvs:
				key = kv[0].strip()
				value = kv[1]
				self.map[key] = value
			return True
		except:
			return False
	def push(self):
		if not self.dirty:
			return
		self.dirty = 0

		fp = open(self.filename)
		file_txt = fp.read()
		fp.close()


		for key in self.map.keys():
			if key == 'filename' or key == 'fullname' or key == 'wikiname':
				continue
			re_key = re.compile(r'\%s(%s)\s*=\s*([^\%s\n]*)\s*\%s' % (prefix, key, suffix, suffix))
			repl = '%s%s=%s%s' % (prefix, key, self.map[key], suffix)
			if re_key.search(file_txt):
				file_txt = re_key.sub(repl, file_txt)
			else:
				file_txt += '\n' + repl + '\n'

		for key in self.del_keys:
			if key == 'filename' or key == 'fullname' or key == 'wikiname':
				continue
			re_key = re.compile(r'\%s(%s)\s*=\s*([^\%s\n]*)\s*\%s' % (prefix, key, suffix, suffix))
			repl = ''
			if re_key.search(file_txt):
				file_txt = re_key.sub(repl, file_txt)

		
		rpc = xmlrpclib.ServerProxy("%s/lib/exe/xmlrpc.php" % (wiki_write_access))
		if wiki_put:
			id = self.attr('wikiname')
			f = open('/tmp/out', 'w')
			f.write(wiki_url)
			f.write('\n')
			f.write(id)
			f.write('\n')
			f.write(file_txt)
			f.write('\n')
			rpc.wiki.putPage(id, file_txt, {})
			f.close()
		else:
			fp = open(self.filename, "w")
			fp.write(file_txt)
			fp.close()

		self.del_keys = []

set = Set()

import lex

tokens = (
	'LPAREN',
	'RPAREN',
	'LOAD',
	'AND',
	'OR',
	'IS',
	'NOT',
	'QWORD',
	'SQWORD',
	'WORD',
	'NUMBER',
	'NEW',
	'DELETE',
	'ATTR',
	'INSET',
	'HAS',
	'MATCH',
	'SORT',
	'SSORT',
	'REV',
	'SET',
	'GT',
	'LT',
	'GTE',
	'LTE',
	'PLUS',
	'CAP'
)

reserved = {
	'and' : 'AND',
	'or' : 'OR',
	'is' : 'IS',
	'not' : 'NOT',
	'load' : 'LOAD',
	'new' : 'NEW',
	'delete' : 'DELETE',
	'has' : 'HAS',
	'match' : 'MATCH',
	'sort' : 'SORT',
	'ssort' : 'SSORT',
	'rev' : 'REV',
	'set' : 'SET',
	'cap' : 'CAP'
	}

t_LPAREN  = r'\('
t_RPAREN  = r'\)'

def t_SP(t):
    r'[ \t\n]'

def t_PLUS(t):
	r'\+'
	return t

def t_INSET(t):
	r'\*'
	return t

def t_GTE(t):
	r'\>\='
	return t

def t_LTE(t):
	r'\<\='
	return t

def t_GT(t):
	r'\>'
	return t

def t_LT(t):
	r'\<'
	return t

def t_WORD(t):
	r'[\./a-zA-Z_][\./a-zA-Z_0-9:]*'
	t.type = reserved.get(t.value,'WORD')
	return t

def t_ATTR(t):
	r'\@[a-zA-Z_][a-zA-Z_0-9]*'
	t.type = reserved.get(t.value,'ATTR')
	return t

def t_QWORD(t):
	r'"[^"]*"'
	t.type = reserved.get(t.value,'QWORD')
	return t

def t_SQWORD(t):
	r"'[^']*'"
	t.type = reserved.get(t.value,'SQWORD')
	return t

def t_NUMBER(t):
	r'\-?(\b[0-9]+\.?([0-9]+\b)?|\.[0-9]+\b)'
	try:
		t.value = str(float(t.value))
	except:
		t.value = str(-1)
	return t

def t_error(t):
	if not suppress_print:
		print "Illegal character in input [%s]" % t.value[0]
	t.skip(1)

t_ignore = ''


precedence = (
    ('left', 'SORT'),
    ('left', 'HAS'),
    ('left', 'AND'),
    ('left', 'OR'),
    ('left', 'NOT', 'IS'),
    ('nonassoc', 'ONLY')
)

lex.lex()

import yacc


def p_set_attr(t):
	'set : set SET attr word'
	t[1].setkv(t[3], t[4])
	t[0] = t[1]

def p_set_num_attr(t):
	'set : set SET attr NUMBER'
	t[1].setkv(t[3], t[4])
	t[0] = t[1]

def p_set_rsort(t):
	'set : set SORT attr REV'
	t[1].sort(t[3], reverse=True)
	t[0] = t[1]

def p_set_sort(t):
	'set : set SORT attr'
	t[1].sort(t[3])
	t[0] = t[1]

def p_set_rssort(t):
	'set : set SSORT attr REV'
	t[1].ssort(t[3], reverse=True)
	t[0] = t[1]

def p_set_ssort(t):
	'set : set SSORT attr'
	t[1].ssort(t[3])
	t[0] = t[1]

def p_set_delete(t):
	'set : set DELETE attr'
	t[1].delete(t[3])
	t[0] = t[1]

def p_set_input(t):
	'set : INSET'
	t[0] = copy.deepcopy(set)

def p_set_load(t):
	'set : LOAD word'
	t[0] = Set()
	t[0].load(t[2])

def p_set_new(t):
	'set : NEW word'
	t[0] = Set()
	t[0].new(t[2])

def p_set_match(t):
	'set : set attr MATCH word'
	t[1].filterHas(t[2], t[4])
	t[0] = t[1]

def p_set_exists(t):
	'set : set HAS attr'
	t[1].filterExists(t[3])
	t[0] = t[1]

def p_set_incr(t):
	'set : set attr PLUS NUMBER'
	t[1].incr(t[2] + ":" + t[4])
	t[0] = t[1]

def p_set_is_num(t):
	'set : set attr IS NUMBER'
	t[1].filterNumEqual(t[2], t[4])
	t[0] = t[1]

def p_set_is(t):
	'set : set attr IS word'
	t[1].filterEqual(t[2], t[4])
	t[0] = t[1]

def p_set_attr_is(t):
	'set : set attr IS attr'
	t[1].filterAttrEqual(t[2], t[4])
	t[0] = t[1]

def p_set_gt(t):
	'set : set attr GT NUMBER'
	t[1].filterGreaterThan(t[2], t[4])
	t[0] = t[1]

def p_set_lt(t):
	'set : set attr LT NUMBER'
	t[1].filterLessThan(t[2], t[4])
	t[0] = t[1]

def p_set_gte(t):
	'set : set attr GTE NUMBER'
	t[1].filterGreaterThanEq(t[2], t[4])
	t[0] = t[1]

def p_set_lte(t):
	'set : set attr LTE NUMBER'
	t[1].filterLessThanEq(t[2], t[4])
	t[0] = t[1]

def p_set_not(t):
	'set : set attr NOT word'
	t[1].filterNotEqual(t[2], t[4])
	t[0] = t[1]

def p_set_cap(t):
	'set : set CAP NUMBER'
	t[1].cap(t[3])
	t[0] = t[1]

def p_word_to_word(t):
	'word : WORD'
	t[0] = t[1]

def p_attr_to_attr(t):
	'attr : ATTR'
	t[0] = t[1][1:]

def p_set_expr(t):
	'set : LPAREN set RPAREN'
	t[0] = t[2]

def p_set_implicit_or(t):
	'set : set set'
	t[1].join(t[2])
	t[0] = t[1]

def p_set_or(t):
	'set : set OR set'
	t[1].join(t[3])
	t[0] = t[1]

def p_set_and(t):
	'set : set AND set'
	t[1].intersect(t[3])
	t[0] = t[1]

def p_sqword_to_word(t):
	'word : SQWORD'
	t[0] = t[1].strip("'")

def p_qword_to_word(t):
	'word : QWORD'
	t[0] = t[1].strip('"')


suppress_print = 0
wiki_put = 0
def p_error(t):
	if not suppress_print:
		print "syntax error in input"
		print t

yacc.yacc(tabmodule='setquerytab')

def handleQuery(query):
	set = yacc.parse(query)
	return set

def handleWikiQuery(query):
	global default_view
	global suppress_print
	global wiki_put
	suppress_print = 1
	wiki_put = 1
	#query = query.lstrip().lstrip('*')
	#query = "* " + query
	default_view = 'html wikiname%s:priority%10.1f:status%s:handler%s:owner%s:submitter%s:title%s'
	try:
		set = yacc.parse(query)
	except:
		set = Set()
	if set == None:
		set = Set()
	return set


instruction_list = []
manual_add = 0

# callbacks
def callback1(option, opt, value, parser):
	global instruction_list
	instruction_list.append((option, value))
	
def callback2(option, opt, value, parser):
	global manual_add
	manual_add = 1
	
default_view = 'fullname%-10s:priority%5.1f: :status%-7s: :handler%-10s:owner%-10s:submitter%-10s: | :title%s'


if __name__ == "__main__":

	user = os.getenv('ISSUE_USER')
	if user == None:
		user = os.getenv('USER')

	# process command line
	parser = optparse.OptionParser()

	# normal state options
	parser.add_option("--terse", dest="terse", action="store_true")

	parser.add_option("-u", "--user", dest="user", default=None,
		help="set USER", metavar='USER')


	# command options (order dependent, operated on the working set)
	W_opt = parser.add_option("-W", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="wiki interface QUERY", metavar='QUERY')

	q_opt = parser.add_option("-q", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="execute QUERY", metavar='QUERY')

	v_opt = parser.add_option("-v", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="view FORMAT", metavar='FORMAT')

	V_opt = parser.add_option("-V", action="callback",
		callback=callback1, nargs=0,
		default=None, help="view")

	L_opt = parser.add_option("-L", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="generate latex report", metavar='TEMPLATE')

	D_opt = parser.add_option("-D", action="callback",
		callback=callback1, nargs=0,
		default=None, help="generate dependencies dot file")


	O_opt = parser.add_option("-O", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="set USER and filter as owner", metavar='USER')

	H_opt = parser.add_option("-H", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="set USER and filter as handler", metavar='USER')

	N_opt = parser.add_option("-N", action="callback",
		callback=callback1, nargs=2, type='string',
		default=None, help="new LOCATION TITLE", metavar='LOCATION TITLE')


	f_opt = parser.add_option("-f", action="callback",
		callback=callback1, nargs=2, type='string',
		default=None, help="filter allowing ATTRIBUTE to be VALUE", metavar='ATTRIBUTE VALUE')

	x_opt = parser.add_option("-x", action="callback",
		callback=callback1, nargs=2, type='string',
		default=None, help="filter disallowing ATTRIBUTE to be VALUE", metavar='ATTRIBUTE VALUE')

	s_opt = parser.add_option("-s", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="sort ATTRIBUTE", metavar='ATTRIBUTE')

	n_opt = parser.add_option("-n", action="callback",
		callback=callback1, nargs=1, type='string',
		default=None, help="new LOCATION", metavar='LOCATION')

	a_opt = parser.add_option("-a", action="callback",
		callback=callback1, nargs=2, type='string',
		default=None, help="assign ATTRIBUTE VALUE", metavar='ATTRIBUTE VALUE')

	i_opt = parser.add_option("-i", action="callback",
		callback=callback1, nargs=0,
		default=None, help="increment priority")

	d_opt = parser.add_option("-d", action="callback",
		callback=callback1, nargs=0,
		default=None, help="decrement priority")

	t_opt = parser.add_option("-t", action="callback",
		callback=callback2, nargs=0,
		default=None, help="text edit body")

	c_opt = parser.add_option("-c", dest='category',
		nargs=1, type='string',
		default=None, help="category")

	I_opt = parser.add_option("-I", action="callback",
		callback=callback1, nargs=1, type='string', metavar='WIKIID',
		default=None, help="set WIKIID")

	g_opt = parser.add_option("-g", action="callback",
		callback=callback1, nargs=1, type='string', metavar='URL',
		default=None, help="get issue files")

	G_opt = parser.add_option("-G", action="callback",
		callback=callback1, nargs=0,
		default=None, help="get issue files")

	p_opt = parser.add_option("-p", action="callback",
		callback=callback1, nargs=0,
		default=None, help="put issue files")



	(options, args) = parser.parse_args()

	
	if options.user:
		user = options.user



	if options.terse == 1:
		set.terse = 1

	if manual_add:
		newset = Set()
		newset.new('.')
		newset.setkv('owner', user)
		newset.setkv('title', ' '.join(args))
		newset.setkv('priority', '0.0')
		newset.setkv('status', 'dormant')
		newset.setkv('submitter', user)
		newset.setkv('category', options.category)
		newset.push()
		newset.edit()
		set = newset
		set.display(default_view)
		sys.exit(0)
		

	if len(args) == 0:
		args = [ '.' ]

	if len(instruction_list) == 0:
		instruction_list.append((x_opt, ('status', 'closed')))
		instruction_list.append((s_opt, 'priority'))
		instruction_list.append((v_opt, default_view))


	for arg in args:
		set.load(arg)

	for instruction in instruction_list:
		if instruction[0] == q_opt:
			set = handleQuery(instruction[1])
		if instruction[0] == I_opt:
			wiki_id = instruction[1]
		if instruction[0] == W_opt:
			set = handleWikiQuery(instruction[1])
		if instruction[0] == v_opt:
			set.display(instruction[1])
		if instruction[0] == V_opt:
			set.display(default_view)
		if instruction[0] == L_opt:
			set.latex(instruction[1])
		if instruction[0] == D_opt:
			set.dotDependencies()
		if instruction[0] == O_opt:
			user = instruction[1]
			set.filterEqual('owner', instruction[1])
			set.filterNotEqual('status', 'closed')
			set.sort('priority', False)
		if instruction[0] == H_opt:
			user = instruction[1]
			set.filterEqual('handler', instruction[1])
			set.filterNotEqual('status', 'closed')
			set.sort('priority', False)
		if instruction[0] == f_opt:
			set.filterEqual(instruction[1][0], instruction[1][1])
		if instruction[0] == x_opt:
			set.filterNotEqual(instruction[1][0], instruction[1][1])
		if instruction[0] == s_opt:
			set.sort(instruction[1], False)
		if instruction[0] == a_opt:
			set.setkv(instruction[1][0], instruction[1][1])
		if instruction[0] == i_opt:
			set.incr('priority:%s' % '1.0')
		if instruction[0] == d_opt:
			set.incr('priority:%s' % '-1.0')
		if instruction[0] == n_opt:
			newset = Set()
			newset.new(instruction[1])
			newset.push()
			set = newset
		if instruction[0] == N_opt:
			newset = Set()
			newset.new(instruction[1][0])
			newset.setkv('owner', user)
			newset.setkv('title', instruction[1][1])
			newset.setkv('priority', '0.0')
			newset.setkv('status', 'dormant')
			newset.setkv('submitter', user)
			newset.push()
			set = newset
			set.display(default_view)
		if instruction[0] == g_opt:
			dirs = {}
			url = instruction[1] + "/lib/exe/xmlrpc.php"
			rpc = xmlrpclib.ServerProxy(url)
			pages = rpc.wiki.getAllPages()
			for page in pages:
				page = page['id']
				fields = page.split(':')
				if fields[0] == 'issue':
					path = ''
					for field in fields[1:]:
						path = os.path.join(path, field)
					(dir,filename) = os.path.split(path)
					if not dirs.has_key(dir):
						dirs[dir] = "URL," + url + "\n"
					if dir != '':
						if not os.path.exists(dir):
							os.makedirs(dir)
					print 'getting [%s]' % (page)
					txt = rpc.wiki.getPage(page)
					info = rpc.wiki.getPageInfo(page)
					m = md5.new()
					m.update(txt)
					dirs[dir] += 'PAGE,' + page + ',' + m.hexdigest() + "," + str(info['lastModified']) + '\n'

					# write page
					fp = open(path, 'w')
					fp.write(txt)
					fp.close()
			for dir in dirs.keys():
				fp_manifest = open(os.path.join(dir,'.issue'), 'w')
				fp_manifest.write(dirs[dir])
				fp_manifest.close()
		if instruction[0] == G_opt:
			urls = []
			known_pages = {}
			for root, dirs, files in os.walk('.'):
				for f in files:
					if f == ".issue":
						fp_manifest = open(os.path.join(root,'.issue'), 'r')
						for line in fp_manifest.readlines():
							line = line.strip()
							fields = line.split(',')
							if fields[0] == 'URL':
								urls.append(fields[1])
							if fields[0] == 'PAGE':
								known_pages[fields[1]] = (fields[2], fields[3])
						fp_manifest.close()

			for url in urls:
				dirs = {}
				rpc = xmlrpclib.ServerProxy(url)
				pages = rpc.wiki.getAllPages()
				for page in pages:
					modified = str(page['lastModified'])
					page = page['id']
					fields = page.split(':')
					if fields[0] == 'issue':
						path = ''
						for field in fields[1:]:
							path = os.path.join(path, field)
						(dir,filename) = os.path.split(path)
						if not dirs.has_key(dir):
							dirs[dir] = "URL," + url + "\n"
						if dir != '':
							if not os.path.exists(dir):
								os.makedirs(dir)

						get_page = 1
						if known_pages.has_key(page):
							if known_pages[page][1] == modified:
								get_page = 0

						if get_page:
							print 'getting [%s]' % (page)
							txt = rpc.wiki.getPage(page)
							info = rpc.wiki.getPageInfo(page)
							m = md5.new()
							m.update(txt)
							dirs[dir] += 'PAGE,' + page + ',' + m.hexdigest() + "," + str(info['lastModified']) + '\n'

							# write page
							fp = open(path, 'w')
							fp.write(txt)
							fp.close()
						else:
							#print 'keeping [%s]' % (page)
							dirs[dir] += 'PAGE,' + page + ',' + known_pages[page][0] + "," + known_pages[page][1] + '\n'
							
				for dir in dirs.keys():
					fp_manifest = open(os.path.join(dir,'.issue'), 'w')
					fp_manifest.write(dirs[dir])
					fp_manifest.close()

								
		if instruction[0] == p_opt:
			url = None
			for root, dirs, files in os.walk('.'):
				for f in files:
					if f == ".issue":
						fp_manifest = open(os.path.join(root,'.issue'), 'r')
						pages = {}
						for line in fp_manifest.readlines():
							line = line.strip()
							fields = line.split(',')
							if fields[0] == 'URL':
								url = fields[1]
							if fields[0] == 'PAGE':
								pages[fields[1]] = fields[2]
						fp_manifest.close()
						if not url:
							continue
						rpc = xmlrpclib.ServerProxy(url)
						dir = root
						if dir == '':
							dir = '.'
						if dir[:2] == "./":
							dir = dir[2:]
						for fn in files:
							if fn[0] == '.':
								continue
							if dir == '.':
								filename = fn
							else:
								filename = os.path.join(dir, fn)
							if not os.path.isfile(filename):
								continue
							fp = open(filename, 'r')
							txt = fp.read()
							fp.close()
							m = md5.new()
							m.update(txt)

							put = 0
							if platform.system()  == "Windows":
								slash = '\\'
							else:
								slash = '/'
							key = 'issue:' + filename.replace(slash, ':');
							if pages.has_key(key):
								if m.hexdigest() != pages[key]:
									put = 1
							else:
								put = 1

							if put:
								print "putting [%s]" % (key)	
								params = {}
								rpc.wiki.putPage(key, txt, params)


	set.push()

	for instruction in instruction_list:
		if instruction[0] == O_opt:
			set.display(default_view)
		if instruction[0] == H_opt:
			set.display(default_view)



