/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "surface/surface.pmh"

#define FE_SAJ_DEBUG		FALSE

namespace fe
{

void SurfaceAccessibleJoint::startJoints(void)
{
	SurfaceAccessibleI::Creation creation=SurfaceAccessibleI::e_createMissing;

	m_spOutputPoint=accessor(e_point,e_position,creation);
	m_spOutputColor=accessor(e_point,"Cd",creation);
	m_spOutputNormal=accessor(e_primitive,"N",creation);
	m_spOutputName=accessor(e_primitive,"name",creation);
	m_spOutputParent=accessor(e_primitive,"parent",creation);
	m_spOutputRefX=accessor(e_primitive,"refX",creation);
	m_spOutputRefY=accessor(e_primitive,"refY",creation);
	m_spOutputRefZ=accessor(e_primitive,"refZ",creation);
	m_spOutputRefT=accessor(e_primitive,"refT",creation);
	m_spOutputAnimX=accessor(e_primitive,"animX",creation);
	m_spOutputAnimY=accessor(e_primitive,"animY",creation);
	m_spOutputAnimZ=accessor(e_primitive,"animZ",creation);
	m_spOutputAnimT=accessor(e_primitive,"animT",creation);
	m_spOutputVertices=accessor(e_primitive,e_vertices,creation);

	m_hiddenTransform.clear();
	m_delegateName.clear();
	m_animMap.clear();
	m_refTweakMap.clear();
	m_animTweakMap.clear();
	m_deltaMap.clear();
}

BWORD SurfaceAccessibleJoint::addJointTweaked(
	String a_name,String a_parentName,
	const SpatialTransform& a_rXformRelative,
	const SpatialTransform& a_rRefWorld,Color a_color)
{
#if FE_SAJ_DEBUG
	feLog("SurfaceAccessibleJoint::addJoint\n");
#endif

	SpatialTransform xformParent;
	SpatialTransform xformHidden;
	String delegate=a_parentName;

	const BWORD hasParent=(m_animMap.find(a_parentName)!=m_animMap.end());
	if(hasParent)
	{
		delegate=m_delegateName[a_parentName];
		xformHidden=m_hiddenTransform[a_parentName];
		xformParent=m_animMap[a_parentName];
	}
	else
	{
		setIdentity(xformParent);
		setIdentity(xformHidden);
	}

	const SpatialTransform xformWorld=a_rXformRelative*xformParent;

	m_animMap[a_name]=xformWorld;

	if(m_spFilter.isValid() && m_spFilter->filter(a_name).empty())
	{
		m_hiddenTransform[a_name]=a_rXformRelative*xformHidden;
		m_delegateName[a_name]=delegate;
		m_refTweakMap[a_name]=a_rRefWorld;
		m_animTweakMap[a_name]=xformWorld;
		setIdentity(m_deltaMap[a_name]);
		return FALSE;
	}

	SpatialTransform xformParentRefTweak;
	SpatialTransform xformParentTweak;
	SpatialTransform deltaParent;

	if(hasParent)
	{
		xformParentRefTweak=m_refTweakMap[a_parentName];
		xformParentTweak=m_animTweakMap[a_parentName];
		deltaParent=m_deltaMap[a_parentName];
	}
	else
	{
		setIdentity(xformParentRefTweak);
		setIdentity(xformParentTweak);
		setIdentity(deltaParent);
	}

	setIdentity(m_hiddenTransform[a_name]);
	m_delegateName[a_name]=a_name;
	SpatialTransform invRefWorld;
	invert(invRefWorld,a_rRefWorld);

	m_deltaMap[a_name]=invRefWorld*xformWorld;

	const SpatialQuaternion quat=a_rXformRelative;

	//* find local tweak in reference

	SpatialTransform invParentRefTweak;
	invert(invParentRefTweak,xformParentRefTweak);

	SpatialTransform refTweak=xformParentRefTweak;

	const SpatialVector toWorld=
			a_rRefWorld.translation()-xformParentRefTweak.translation();
	const SpatialVector toLocal=rotateVector(invParentRefTweak,toWorld);
	if(magnitudeSquared(toLocal)>1e-6)
	{
		const SpatialVector fromLocal(0.0,1.0,0.0);

		SpatialQuaternion quatLocal;
		set(quatLocal,fromLocal,unitSafe(toLocal));

		const SpatialTransform localTweak=quatLocal;

		refTweak=localTweak*refTweak;
	}
	setTranslation(refTweak,a_rRefWorld.translation());

	SpatialTransform animTweak=refTweak*deltaParent;
	setTranslation(animTweak,xformWorld.translation());

	m_refTweakMap[a_name]=refTweak;
	m_animTweakMap[a_name]=animTweak;

	SpatialTransform invParentTweak;
	invert(invParentTweak,xformParentTweak);

	SpatialTransform xformLocal=animTweak*invParentTweak;
	SpatialTransform xformLocalRef=refTweak*invParentRefTweak;

	return addJoint(a_name,a_parentName,xformLocal,animTweak,
			xformLocalRef,a_color);
}

BWORD SurfaceAccessibleJoint::addJoint(String a_name,String a_parentName,
	const SpatialTransform& a_rAnimLocal,
	const SpatialTransform& a_rAnimWorld,
	const SpatialTransform& a_rRefLocal,
	Color a_color)
{
	const BWORD success=addJoint(a_name,a_parentName,
			a_rAnimLocal,a_rAnimWorld,a_color);

	if(success)
	{
		const I32 primitiveIndex=m_spOutputVertices->count()-1;

		m_spOutputRefX->set(primitiveIndex,a_rRefLocal.column(0));
		m_spOutputRefY->set(primitiveIndex,a_rRefLocal.column(1));
		m_spOutputRefZ->set(primitiveIndex,a_rRefLocal.column(2));
		m_spOutputRefT->set(primitiveIndex,a_rRefLocal.translation());
	}

	return success;
}

BWORD SurfaceAccessibleJoint::addJoint(String a_name,String a_parentName,
	const SpatialTransform& a_rAnimLocal,
	const SpatialTransform& a_rAnimWorld,
	Color a_color)
{
	//* TODO param
	const Real triScale=1.0;

	const Color red(1.0,0.0,0.0);
	const Color green(0.0,1.0,0.0);

	const I32 primitiveIndex=m_spOutputVertices->append(3);
	m_spOutputName->set(primitiveIndex,a_name);
	m_spOutputParent->set(primitiveIndex,a_parentName);
	m_spOutputNormal->set(primitiveIndex,a_rAnimWorld.column(1));

	const I32 pointIndex0=m_spOutputVertices->integer(primitiveIndex,0);
	const I32 pointIndex1=m_spOutputVertices->integer(primitiveIndex,1);
	const I32 pointIndex2=m_spOutputVertices->integer(primitiveIndex,2);

	const SpatialVector point=a_rAnimWorld.translation();
	m_spOutputPoint->set(pointIndex0,point);
	m_spOutputPoint->set(pointIndex1,point+triScale*a_rAnimWorld.column(0));
	m_spOutputPoint->set(pointIndex2,point+triScale*a_rAnimWorld.column(2));

	m_spOutputColor->set(pointIndex0,a_color);
	m_spOutputColor->set(pointIndex1,red);
	m_spOutputColor->set(pointIndex2,green);

	m_spOutputAnimX->set(primitiveIndex,a_rAnimLocal.column(0));
	m_spOutputAnimY->set(primitiveIndex,a_rAnimLocal.column(1));
	m_spOutputAnimZ->set(primitiveIndex,a_rAnimLocal.column(2));
	m_spOutputAnimT->set(primitiveIndex,a_rAnimLocal.translation());

	return TRUE;
}

} // namespace
