/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "core.pmh"

namespace fe
{

String InfoI32::print(void *instance)
{
	I32 i32= *(I32 *)instance;
	String string;
	string.sPrintf("%d",i32);
	return string;
}

IWORD InfoI32::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	if(mode == e_binary)
	{
		I32 i32 = htonl(*(I32 *)instance);
		ostrm.write((char *)&i32, sizeof(I32));
		return sizeof(I32);
	}
	else
	{
		I32 i32= *(I32 *)instance;
		ostrm << i32;
		return c_ascii;
	}
}

void InfoI32::input(std::istream &istrm, void *instance, t_serialMode mode)
{
	I32 i32;
	if(mode == e_binary)
	{
		istrm.read((char *)&i32, sizeof(I32));
		*(I32 *)instance = ntohl(i32);
	}
	else
	{
		istrm >> i32;
		*(I32 *)instance = i32;
	}
}

void InfoI32::construct(void *instance)
{
	I32 i32= *(I32 *)instance;
	i32=0;
}

IWORD InfoI32::iosize(void)
{
	return sizeof(I32);
}

String InfoIWORD::print(void *instance)
{
	IWORD iword= *(IWORD *)instance;
	String string;
	string.sPrintf("%d",iword);
	return string;
}

IWORD InfoIWORD::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	if(mode == e_binary)
	{
		IWORD iword = htonl(*(IWORD *)instance);
		ostrm.write((char *)&iword, sizeof(IWORD));
		return sizeof(IWORD);
	}
	else
	{
		IWORD iword= *(IWORD *)instance;
		ostrm << iword;
		return c_ascii;
	}
}

void InfoIWORD::input(std::istream &istrm, void *instance, t_serialMode mode)
{
	IWORD iword;
	if(mode == e_binary)
	{
		istrm.read((char *)&iword, sizeof(IWORD));
		*(IWORD *)instance = ntohl(iword);
	}
	else
	{
		istrm >> iword;
		*(IWORD *)instance = iword;
	}
}

void InfoIWORD::construct(void *instance)
{
	IWORD iword= *(IWORD *)instance;
	iword=0;
}

IWORD InfoIWORD::iosize(void)
{
	return sizeof(IWORD);
}

// ----------------------------------------------------------------------

String InfoF32::print(void *instance)
{
	F32 f32= *(F32 *)instance;
	String string;
	string.sPrintf("%.6G",f32);
	return string;
}

IWORD InfoF32::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	F32 f32= *(F32 *)instance;
	if(mode == e_binary)
	{
		ostrm.write((char *)&f32, sizeof(F32));
		return sizeof(F32);
	}
	else
	{
		ostrm << f32;
		return c_ascii;
	}
}

void InfoF32::input(std::istream &istrm, void *instance, t_serialMode mode)
{
	F32 f32=0.0f;
	if(mode == e_binary)
	{
		istrm.read((char *)&f32, sizeof(F32));
	}
	else
	{
		istrm >> f32;
	}
	*(F32 *)instance = f32;
}

void InfoF32::construct(void *instance)
{
	F32 f32= *(F32 *)instance;
	f32=0.0f;
}

IWORD InfoF32::iosize(void)
{
	return sizeof(F32);
}

// ----------------------------------------------------------------------

String InfoF64::print(void *instance)
{
	F64 f64= *(F64 *)instance;
	String string;
	string.sPrintf("%.6G",f64);
	return string;
}

IWORD InfoF64::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	F64 f64= *(F64 *)instance;
	if(mode == e_binary)
	{
		ostrm.write((char *)&f64, sizeof(F64));
		return sizeof(F64);
	}
	else
	{
		ostrm << f64;
		return c_ascii;
	}
}

void InfoF64::input(std::istream &istrm, void *instance, t_serialMode mode)
{
	F64 f64=0.0f;
	if(mode == e_binary)
	{
		istrm.read((char *)&f64, sizeof(F64));
	}
	else
	{
		istrm >> f64;
	}
	*(F64 *)instance = f64;
}

void InfoF64::construct(void *instance)
{
	F64 f64= *(F64 *)instance;
	f64=0.0f;
}

IWORD InfoF64::iosize(void)
{
	return sizeof(F64);
}

// ----------------------------------------------------------------------

class InfoString : public TypeInfoConstructable<String>
{
	public:
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode)
				{
					if(mode == e_binary)
					{
						String string= *(String *)instance;
						I32 n = htonl(string.length());
						ostrm.write((char *)&n, sizeof(I32));
						ostrm.write(string.c_str(), string.length());
						return sizeof(I32) + string.length();
					}
					else
					{
						String string= *(String *)instance;
						ostrm << "\"" << string.c_str() << "\"";
						return c_ascii;
					}
				}
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode)
				{
					if(mode == e_binary)
					{
						I32 n;
						istrm.read((char *)&n, sizeof(I32));
						n = ntohl(n);
						char buffer[FESTRING_LOCALSIZE];
						char *ptr;
						if(n > FESTRING_LOCALSIZE)
						{
							ptr = new char[n+1];
						}
						else
						{
							ptr = buffer;
						}
						istrm.read(ptr, n);
						ptr[n] = '\0';
						*(String *)instance = ptr;
						if(ptr != buffer) { delete [] ptr; }
					}
					else
					{
						std::string s;
						char c;
						while(istrm.get(c))
						{
							if(c == '\0')
							{
								break;
							}
							s.push_back(c);
						}
						*(String *)instance = s.c_str();
					}
				}
virtual	BWORD	copy(void *instance,const void* source)
				{
					String& rInstance=*(String *)instance;
					String& rSource=*(String *)source;
					rInstance=rSource;
					return TRUE;
				}
virtual	String	print(void *instance)
				{	return *(String *)instance; }
virtual	IWORD	iosize(void)
				{	return FESTRING_LOCALSIZE; }
};

typedef std::map<String,String> t_mapStrStr;
class FE_DL_EXPORT InfoStringStringMap:
		public TypeInfoConstructable<t_mapStrStr>
{
	public:
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode)
				{
					t_mapStrStr &map = *(t_mapStrStr *)instance;

					if(mode == e_binary)
					{
						U32 bytes=0;
						I32 size = htonl(map.size());
						ostrm.write((char *)&size, sizeof(I32));
						bytes+=sizeof(I32);

						t_mapStrStr::const_iterator it=map.begin();
						while(it!=map.end())
						{
							String key=it->first;
							size = htonl(key.length());
							ostrm.write((char *)&size, sizeof(I32));
							bytes+=sizeof(I32);

							ostrm.write(key.c_str(), key.length());
							bytes+=key.length();

							String value=it->second;
							size = htonl(value.length());
							ostrm.write((char *)&size, sizeof(I32));
							bytes+=sizeof(I32);

							ostrm.write(value.c_str(), value.length());
							bytes+=value.length();

							it++;
						}

						return bytes;
					}
					else
					{
						t_mapStrStr::const_iterator it=map.begin();
						if(it==map.end())
						{
							ostrm << "\"\"";
							return c_ascii;
						}
						ostrm << "\n\t\"\n";
						while(it!=map.end())
						{
							ostrm << "\t";
							ostrm << it->first.maybeQuote("'").c_str();
							ostrm << " ";
							ostrm << it->second.maybeQuote("'").c_str();
							ostrm << "\n";

							it++;
						}
						ostrm << "\t\"";

						return c_ascii;
					}
				}
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode)
				{
					t_mapStrStr &map = *(t_mapStrStr *)instance;

					if(mode == e_binary)
					{
						I32 size;
						istrm.read((char *)&size, sizeof(I32));
						size = ntohl(size);

						for(I32 index=0;index<size;index++)
						{
							String key;
							for(U32 pass=0;pass<2;pass++)
							{
								I32 n;
								istrm.read((char *)&n, sizeof(I32));
								n = ntohl(n);
								char buffer[FESTRING_LOCALSIZE];
								char *ptr;
								if(n > FESTRING_LOCALSIZE)
								{
									ptr = new char[n+1];
								}
								else
								{
									ptr = buffer;
								}
								istrm.read(ptr, n);
								ptr[n] = '\0';

								if(pass)
								{
									map[key]=ptr;
								}
								else
								{
									key=ptr;
								}
								if(ptr != buffer)
								{
									delete [] ptr;
								}
							}
						}
					}
					else
					{
						std::string s;
						char c;
						while(istrm.get(c))
						{
							if(c == '\0')
							{
								break;
							}
							s.push_back(c);
						}
						String buffer = s.c_str();

						while(!buffer.empty())
						{
							String key=buffer.parse("'");
							if(key.empty())
							{
								break;
							}
							String value=buffer.parse("'");
							map[key]=value;
						}
					}
				}
virtual	String	print(void *instance)
				{
					String s = "";
					t_mapStrStr &map = *(t_mapStrStr *)instance;

					t_mapStrStr::const_iterator it=map.begin();
					while(it!=map.end())
					{
						s.cat(it->first);
						s.cat(" ");
						s.cat(it->second);
						s.cat("\n");

						it++;
					}

					return s;
				}
virtual	IWORD	iosize(void)
				{
					return BaseType::Info::c_implicit;
				}
};

typedef std::vector<String> t_vecstring;
class FE_DL_EXPORT InfoStringArray : public TypeInfoConstructable<t_vecstring>
{
	public:
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode)
				{
					t_vecstring &v = *(t_vecstring *)instance;
					if(mode == e_binary)
					{
						U32 bytes=0;
						long cnt = (long)v.size();

						I32 n = htonl(cnt);
						ostrm.write((char *)&n, sizeof(I32));
						bytes+=sizeof(I32);

						for(long i = 0; i < cnt; i++)
						{
							String word=v[i];
							I32 len = htonl(word.length());
							ostrm.write((char *)&len, sizeof(I32));
							bytes+=sizeof(I32);

							ostrm.write(word.c_str(), word.length());
							bytes+=word.length();
						}
						return bytes;
					}
					else
					{
						long cnt = (long)v.size();
						if(!cnt)
						{
							ostrm << "\"\"";
							return c_ascii;
						}

						ostrm << "\n\t\"\n";
						for(long i = 0; i < cnt; i++)
						{
							ostrm << "\t";
							ostrm << v[i].maybeQuote("'").c_str();
							ostrm << "\n";
						}
						ostrm << "\t\"";

						return c_ascii;
					}
				}
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode)
				{
					t_vecstring &v = *(t_vecstring *)instance;

					if(mode == e_binary)
					{
						I32 cnt;
						istrm.read((char *)&cnt, sizeof(I32));
						cnt = ntohl(cnt);

						v.resize(cnt);
						for(I32 i = 0; i < cnt; i++)
						{
							I32 len;
							istrm.read((char *)&len, sizeof(I32));
							len = ntohl(len);
							char buffer[FESTRING_LOCALSIZE];
							char *ptr;
							if(len > FESTRING_LOCALSIZE)
							{
								ptr = new char[len+1];
							}
							else
							{
								ptr = buffer;
							}
							istrm.read(ptr, len);
							ptr[len] = '\0';

							v[i]=ptr;

							if(ptr != buffer)
							{
								delete[] ptr;
							}
						}
					}
					else
					{
						std::string s;
						char c;
						while(istrm.get(c))
						{
							if(c == '\0')
							{
								break;
							}
							s.push_back(c);
						}
						String buffer = s.c_str();

						unsigned int cnt=0;
						while(!buffer.empty())
						{
							String word=buffer.parse("'");
							if(!word.empty())
							{
								v.resize(cnt+1);
								v[cnt++]=word;
							}
						}
					}
				}
virtual	String	print(void *instance)
				{
					String s = "";
					t_vecstring &v = *(t_vecstring *)instance;
					for(unsigned int i = 0; i < v.size(); i++)
					{
						if(i > 0)
						{
							s.cat(" ");
						}
						s.cat("%s", v[i].c_str());
					}
					return s;
				}
virtual	IWORD	iosize(void)
				{
					return BaseType::Info::c_implicit;
				}
};

class FE_DL_EXPORT InfoIntegerArray : public TypeInfoConstructable<t_vecint>
{
	public:
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode)
				{
					if(mode == e_binary)
					{
						t_vecint &v = *(t_vecint *)instance;
						long cnt = (long)v.size();
						I32 n = htonl(cnt);

						ostrm.write((char *)&n, sizeof(I32));
						for(long i = 0; i < cnt; i++)
						{
							I32 integer = (I32)v[i];
							ostrm.write((char *)&integer, sizeof(I32));
						}
						return sizeof(I32) + cnt * sizeof(I32);
					}
					else
					{
						t_vecint &v = *(t_vecint *)instance;
						long cnt = (long)v.size();
						ostrm << cnt;
						for(long i = 0; i < cnt; i++)
						{
							ostrm << ",";
							ostrm << v[i];
						}
						return c_ascii;
					}
				}
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode)
				{
					if(mode == e_binary)
					{
						I32 n;
						istrm.read((char *)&n, sizeof(I32));
						n = ntohl(n);

						t_vecint &v = *(t_vecint *)instance;
						v.resize(n);
						for(I32 i = 0; i < n; i++)
						{
							I32 integer;
							istrm.read((char *)&integer, sizeof(I32));
							v[i] = integer;
						}
					}
					else
					{
						I32 n;
						istrm >> n;
						t_vecint &v = *(t_vecint *)instance;
						v.resize(n);
						for(I32 i = 0; i < n; i++)
						{
							char comma;
							istrm >> comma;
							I32 integer;
							istrm >> integer;
							v[i] = integer;
						}
					}
				}
virtual	String	print(void *instance)
				{
					String s = "";
					t_vecint &v = *(t_vecint *)instance;
					for(unsigned int i = 0; i < v.size(); i++)
					{
						if(i > 0)
						{
							s.cat(" ");
						}
						s.cat("%d", v[i]);
					}
					return s;
				}
virtual	IWORD	iosize(void)
				{
					return BaseType::Info::c_implicit;
				}
};

class FE_DL_EXPORT InfoRealArray : public TypeInfoConstructable<t_vecreal>
{
	public:
virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode)
				{
					if(mode == e_binary)
					{
						t_vecreal &v = *(t_vecreal *)instance;
						long cnt = (long)v.size();
						I32 n = htonl(cnt);

						ostrm.write((char *)&n, sizeof(I32));
						for(long i = 0; i < cnt; i++)
						{
							F64 f64 = (F64)v[i];
							ostrm.write((char *)&f64, sizeof(F64));
						}
						return sizeof(I32) + cnt * sizeof(F64);
					}
					else
					{
						t_vecreal &v = *(t_vecreal *)instance;
						long cnt = (long)v.size();
						ostrm << cnt;
						for(long i = 0; i < cnt; i++)
						{
							ostrm << ",";
							ostrm << v[i];
						}
						return c_ascii;
					}
				}
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode)
				{
					if(mode == e_binary)
					{
						I32 n;
						istrm.read((char *)&n, sizeof(I32));
						n = ntohl(n);

						t_vecreal &v = *(t_vecreal *)instance;
						v.resize(n);
						for(I32 i = 0; i < n; i++)
						{
							F64 f64;
							istrm.read((char *)&f64, sizeof(F64));
							v[i] = f64;
						}
					}
					else
					{
						I32 n;
						istrm >> n;
						t_vecreal &v = *(t_vecreal *)instance;
						v.resize(n);
						for(I32 i = 0; i < n; i++)
						{
							char comma;
							istrm >> comma;
							F64 f64;
							istrm >> f64;
							v[i] = f64;
						}
					}
				}
virtual	String	print(void *instance)
				{
					String s = "";
					t_vecreal &v = *(t_vecreal *)instance;
					for(unsigned int i = 0; i < v.size(); i++)
					{
						if(i > 0)
						{
							s.cat(" ");
						}
						s.cat("%f", v[i]);
					}
					return s;
				}
virtual	IWORD	iosize(void)
				{
					return BaseType::Info::c_implicit;
				}
};

class FE_DL_EXPORT InfoIntIntArray :
	public TypeInfoConstructable< std::vector< std::vector<I32> > >
{
	public:
				//* TODO

virtual	IWORD	output(std::ostream &ostrm, void *instance, t_serialMode mode)
				{
				}
virtual	void	input(std::istream &istrm, void *instance, t_serialMode mode)
				{
				}
virtual	String	print(void *instance)
				{
				}
virtual	IWORD	iosize(void)
				{
					return BaseType::Info::c_implicit;
				}
};

String InfoBool::print(void *instance)
{
	bool b= *(bool *)instance;
	String string;
	if(b)
	{
		string.sPrintf("true");
	}
	else
	{
		string.sPrintf("false");
	}
	return string;
}

IWORD InfoBool::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	if(mode == e_binary)
	{
		U8 b;
		if(*(bool *)instance)
		{
			b = 1;
		}
		else
		{
			b = 0;
		}
		ostrm.write((char *)&b, sizeof(U8));
		return sizeof(U8);
	}
	else
	{
		bool b= *(bool *)instance;
		ostrm << b;
		return c_ascii;
	}
}

void InfoBool::input(std::istream &istrm, void *instance, t_serialMode mode)
{
	if(mode == e_binary)
	{
		U8 u8;
		istrm.read((char *)&u8, sizeof(U8));
		*(bool *)instance = (u8==1);
	}
	else
	{
		std::string s;
		istrm >> s;

		if (s == "0" || s == "false" || s == "FALSE" || s == "False")
		{
			*(bool *)instance = false;
		}
		else
		{
			*(bool *)instance = true;
		}
	}
}

void InfoBool::construct(void *instance)
{
	bool b= *(bool *)instance;
	b=false;
}

IWORD InfoBool::iosize(void)
{
	return sizeof(U8);
}

String InfoVoid::print(void *instance)
{
	String string;
	return string;
}

IWORD InfoVoid::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	return 0;
}

void InfoVoid::input(std::istream &istrm, void *instance, t_serialMode mode)
{
}

void InfoVoid::construct(void *instance)
{
}

IWORD InfoVoid::iosize(void)
{
	return 0;
}

String InfoVoidStar::print(void *instance)
{
	String string;
	return string;
}

IWORD InfoVoidStar::output(std::ostream &ostrm, void *instance, t_serialMode mode)
{
	return 0;
}

void InfoVoidStar::input(std::istream &istrm, void *instance, t_serialMode mode)
{
}

void InfoVoidStar::construct(void *instance)
{
}

IWORD InfoVoidStar::iosize(void)
{
	return 0;
}

void assertCore(sp<TypeMaster> spTypeMaster)
{
	sp<BaseType> spT;
	sp<BaseType::Info> spInfo;

	spT = spTypeMaster->assertType<I32>("I32");
	spT->setInfo(new InfoI32());
	spT = spTypeMaster->assertType<IWORD>("IWORD");
	spT->setInfo(new InfoIWORD());
	spT = spTypeMaster->assertType<F32>("F32");
	spT->setInfo(new InfoF32());
	spT = spTypeMaster->assertType<F64>("F64");
	spT->setInfo(new InfoF64());
	spT = spTypeMaster->assertType<String>("string");
	spT->setInfo(new InfoString());
	spT = spTypeMaster->assertType<void>("void");
	spT->setInfo(new InfoVoid());
	spT = spTypeMaster->assertType<void*>("voidstar");
	spT->setInfo(new InfoVoidStar());

	spT = spTypeMaster->assertType< std::map<String,String> >(
			"stringstringmap");
	spT->setInfo(new InfoStringStringMap());

	spT = spTypeMaster->assertType< std::vector<String> >("stringarray");
	spT->setInfo(new InfoStringArray());

	spT = spTypeMaster->assertType< std::vector<I32> >("intarray");
	spT->setInfo(new InfoIntegerArray());

	spT = spTypeMaster->assertType< std::vector<Real> >("realarray");
	spT->setInfo(new InfoRealArray());

	spT = spTypeMaster->assertType< std::vector< std::vector<I32> > >(
			"intintarray");
	spT->setInfo(new InfoIntIntArray());

	// platform and language neutral names
	spT = spTypeMaster->assertType<I32>("integer");
	spT = spTypeMaster->assertType<Real>("real");
	spT = spTypeMaster->assertType<bool>("boolean");
	spT->setInfo(new InfoBool());

	// C/C++ type mappings
	spT = spTypeMaster->assertType<int>("int");
	spT = spTypeMaster->assertType<float>("float");
	spT = spTypeMaster->assertType<double>("double");
	spT = spTypeMaster->assertType<bool>("bool");


	// not sure we want this:
	//spT = spTypeMaster->assertType<Instance>("any");
}

} /* namespace */

